# This module is part of the license usage graph package. It requires
# the GD module, which can be installed on ActivePerl on Win32 with PPM
# from the Winnipeg repository: http://theoryx5.uwinnipeg.ca/ppms
# On Unix platorms, along with the GD module, you must install libgd
# and libpng according to the requirements of the GD module.

package GraphLicenses;

use GD;
use Time::Local;

sub new {
    my ($class, $type) = @_;
    my $objref = initialize($type);
    return undef if not $objref;
    bless $objref, $class;
    return $objref;
}

sub graph {
    my ($s, $infile) = @_;
    $s->load_license_samples($infile);
    $s->initialize_image;
    $s->calculate_scale_values;
    $s->add_title_to_graph;
    $s->add_legend_to_graph;
    $s->label_x_axis;
    $s->label_y_axis;
    $s->generate_graph;
    $s->add_highest_usage_to_graph;
    return $s;
}

sub save {
    my ($s, $graph_file) = @_;

    # Write the image
    open PNG, "> $graph_file" or die "Can't open $graph_file: $!\n";
    binmode PNG;
    print PNG $s->{image}->png(0);
    close PNG;
}

#
# Mark the highest usage
#
sub add_highest_usage_to_graph($) {
    my $s = shift;

    # Graph the "highest ever" usage
    my $start_x =
      $s->{x_origin} +
      (($s->{sample}->{min_sample_time} - $s->{x_start_time}) / $s->{x_scale});
    my $end_x =
      $s->{x_origin} +
      (($s->{sample}->{max_sample_time} - $s->{x_start_time}) / $s->{x_scale});
    my $y_val =
      ($s->{height} - $s->{bottom_margin}) - ($s->{sample}->{max_used_ever} / $s->{y_scale});

    $s->{image}->dashedLine($start_x, $y_val, $end_x, $y_val, $s->{max_used_ever_color});
}

#
# Put a legend/key on the graph. Strings are right-justified
#
sub add_legend_to_graph($) {
    my $s = shift;

    my $y_val = $s->{top_margin} / 2;
    my ($text, $title_x);

    $text = sprintf("Available now: %3d", $s->{sample}->{max_licenses_avail});
    $title_x = $s->{width} - $s->{right_margin} - (length($text) * $s->{legend_font}->width);
    $s->{image}->string($s->{legend_font}, $title_x, $y_val + ($s->{legend_font}->height * 0),
        $text, $s->{max_licenses_avail_color});

    $text = sprintf("Highest ever: %3d", $s->{sample}->{max_used_ever});
    $title_x = $s->{width} - $s->{right_margin} - (length($text) * $s->{legend_font}->width);
    $s->{image}->string($s->{legend_font}, $title_x, $y_val + ($s->{legend_font}->height * 1),
        $text, $s->{max_used_ever_color});

    $text = sprintf("Highest this period: %3d", $s->{sample}->{max_used_this_period});
    $title_x = $s->{width} - $s->{right_margin} - (length($text) * $s->{legend_font}->width);
    $s->{image}->string($s->{legend_font}, $title_x, $y_val + ($s->{legend_font}->height * 2),
        $text, $s->{max_used_this_period_color});

}

#
# Put a title on the graph
#
sub add_title_to_graph($) {
    my $s = shift;

    my (@start_date, @start_time, @end_date, $text);

    # Create the title
    if ($s->{type} eq 'today') {
        @start_date = split /\s+/, localtime;

        $text = "";
        @start_time = split /:/, $start_date[3];
        $text .=
          "$start_date[1] $start_date[2], $start_date[4] at $start_time[0]:$start_time[1]";
    }
    else {
        @start_date = split /\s+/, localtime($s->{x_start_time});
        @end_date   = split /\s+/, localtime($s->{x_end_time});

        $text = "";
        $text .=
            "$start_date[1] $start_date[2], $start_date[4] "
          . "thru $end_date[1] $end_date[2], $end_date[4]";
    }

    # Center the title.
    my $title_x = 0;
    $s->{image}
      ->string($s->{title_font}, $title_x, ($s->{top_margin} - $s->{title_font}->height) / 2,
        $text, $s->{title_color});
}

#
# Calculate scaling factors.
#
sub calculate_scale_values($) {
    my $s = shift;

    my $sample = $s->{sample};

    $s->{x_scale} =
      ($s->{x_end_time} - $s->{x_start_time}) /
      ($s->{width} - $s->{left_margin} - $s->{right_margin});

    # Subtract the 4 lines for the legend. This prevents it from being
    # overwritten.

    # Warning: Sloppy code ahead! We want 5 hash marks on the y-axis,
    # but they all should be whole numbers. Round the max. up until
    # it works.
    $s->{max_y_val} = $sample->{max_licenses_avail};
    $s->{max_y_val} += 5 - ($s->{max_y_val} % 5) if ($s->{max_y_val} % 5);

    $s->{y_scale} =
      ($s->{max_y_val} - 0) /
      ($s->{height} - $s->{top_margin} - $s->{bottom_margin} - ($s->{legend_font}->height * 2)
      );
}

#
# Generate the graph
#
sub generate_graph($) {
    my $s = shift;

    my ($prev_x_val, $prev_y_val, $prev_y_avail) = (-1);
    for my $sample_time (sort keys %{$s->{sample}->{avail}}) {

        # The x and y-coordinates of the data point
        my $x_val = $s->{x_origin} + (($sample_time - $s->{x_start_time}) / $s->{x_scale});
        my $y_val =
          ($s->{height} - $s->{bottom_margin}) -
          (${$s->{sample}->{used}}{$sample_time} / $s->{y_scale});

        # The y-coordinate of the "available" point
        my $y_avail =
          ($s->{height} - $s->{bottom_margin}) -
          (${$s->{sample}->{avail}}{$sample_time} / $s->{y_scale});

        if ($s->{style} eq "points") {

            # Plot the data point
            $s->{image}->setPixel($x_val, $y_val, $s->{usage_color});

            # Plot the "available" point
            $s->{image}->setPixel($x_val, $y_avail, $s->{max_licenses_avail_color});
        }
        elsif ($s->{style} eq "lines") {
            if ($prev_x_val != -1) {

                # Plot the data point
                $s->{image}->line($prev_x_val, $prev_y_val, $x_val, $y_val, $s->{usage_color});

                # Plot the "available" point
                $s->{image}->line($prev_x_val, $prev_y_avail, $x_val, $y_avail,
                    $s->{max_licenses_avail_color});
            }
            $prev_x_val   = $x_val;
            $prev_y_val   = $y_val;
            $prev_y_avail = $y_avail;

        }
        elsif ($s->{style} eq "impulses") {

            # Plot the data point
            if (${$s->{sample}->{used}}{$sample_time} > 0) {

                # Don't draw on the x-axis
                $s->{image}->line($x_val, $s->{height} - $s->{bottom_margin} - 1,
                    $x_val, $y_val, $s->{usage_color});

            }
            else {
                $s->{image}->line($x_val, $s->{height} - $s->{bottom_margin},
                    $x_val, $y_val, $s->{usage_color});

            }

            # Plot the "available" point
            $s->{image}->setPixel($x_val, $y_avail, $s->{max_licenses_avail_color});
        }

        # If this point is (one of) the highest for this period,
        # mark it with a diamond.
        # Draw from west of the point, clockwise.
        if (    ${$s->{sample}->{used}}{$sample_time}
            and ${$s->{sample}->{used}}{$sample_time} ==
            $s->{sample}->{max_used_this_period})
        {
            my $poly = new GD::Polygon;
            $poly->addPt($x_val - 3, $y_val);
            $poly->addPt($x_val,     $y_val - 3);
            $poly->addPt($x_val + 3, $y_val);
            $poly->addPt($x_val,     $y_val + 3);
            $s->{image}->polygon($poly, $s->{max_used_this_period_color});
        }
    }
}

#
# Initialize stuff
#
sub initialize($) {
    my $type = shift;

    my $values = {
        type => $type,

        width  => 600,
        height => 225,

        left_margin   => 40,
        right_margin  => 40,
        top_margin    => 40,
        bottom_margin => 40,

        # The length of a hash mark
        hash_length => 4,

        # Coordinates of the last point plotted
        prev_x_val => -1,
        prev_y_val => -1,

        # keep track of maximum license usages
        max_used_this_period => 0,
        max_used_ever        => 0,
        max_licenses_avail   => 0,

        max_sample_time => 0,
        min_sample_time => 0
    };

    # Coordinates of the origin
    $values->{x_origin} = $values->{left_margin};
    $values->{y_origin} = $values->{height} - $values->{bottom_margin};

    # calculate the beginning x value for the graph. initally, set it to
    # 12am today
    my ($day, $month, $year) = (localtime)[ 3, 4, 5 ];
    $values->{x_start_time} = timelocal(0, 0, 0, $day, $month, $year);

    # 86400 is the number of seconds in a day
    # go back in time, if necessary
    if ($type eq 'today') {
        $values->{seconds_on_graph} = 86400;
        $values->{style}            = 'lines';
        $values->{x_hour_marks}     = 3;
    }
    elsif ($type eq 'week') {
        $values->{seconds_on_graph} = 86400 * 7;
        $values->{x_start_time} -= 86400 * 6;
        $values->{style} = 'lines';
    }
    elsif ($type eq 'month') {
        $values->{seconds_on_graph} = 86400 * 30;
        $values->{x_start_time} -= 86400 * 29;
        $values->{style} = 'lines';
    }
    elsif ($type eq 'year') {
        $values->{seconds_on_graph} = 86400 * 365;
        $values->{x_start_time} -= 86400 * 364;
        $values->{style} = 'impulses';
    }
    else {
        return undef;
    }

    # the end position is always midnight tonight
    $values->{x_end_time} = $values->{x_start_time} + $values->{seconds_on_graph} - 1;

    return $values;
}

#
# Initialize the image
#
sub initialize_image($) {
    my $s = shift;

    # Create a new image
    my $image = new GD::Image($s->{width}, $s->{height});

    # Allocate colors. Values are from the "rgbcolors" command
    my $black        = $image->colorAllocate(0,   0,   0);
    my $blue         = $image->colorAllocate(0,   0,   255);
    my $brown        = $image->colorAllocate(165, 42,  42);
    my $green        = $image->colorAllocate(0,   235, 12);
    my $grey         = $image->colorAllocate(180, 180, 180);
    my $dark_grey    = $image->colorAllocate(90,  90,  90);
    my $light_grey   = $image->colorAllocate(194, 194, 194);
    my $medium_green = $image->colorAllocate(0,   166, 33);
    my $navy_blue    = $image->colorAllocate(0,   0,   80);
    my $olive        = $image->colorAllocate(107, 142, 35);
    my $purple       = $image->colorAllocate(160, 32,  240);
    my $red          = $image->colorAllocate(255, 0,   0);
    my $violet       = $image->colorAllocate(255, 0,   255);
    my $white        = $image->colorAllocate(255, 255, 255);
    my $yellow       = $image->colorAllocate(255, 255, 0);

    #====================================
    # Configurable parameters start here
    #====================================

    $s->{axis_color}       = $black;
    $s->{background_color} = $white;
    $s->{title_color}      = $black;
    $s->{usage_color}      = $blue;

    $s->{max_licenses_avail_color}   = $red;
    $s->{max_used_ever_color}        = $purple;
    $s->{max_used_this_period_color} = $black;

    $s->{label_font}  = GD::Font->Small;
    $s->{legend_font} = GD::Font->Small;
    $s->{title_font}  = GD::Font->Large;

    #====================================
    # End of configurable parameters
    #====================================

    # Set the background
    $image->filledRectangle(0, 0, $s->{width}, $s->{height}, $s->{background_color});

    # Draw the left y-axis
    $image->line($s->{x_origin}, $s->{top_margin}, $s->{x_origin},
        $s->{height} - $s->{bottom_margin},
        $s->{axis_color});

    # Draw the bottom x-axis
    $image->line(
        $s->{x_origin},
        $s->{height} - $s->{bottom_margin},
        $s->{width} - $s->{right_margin},
        $s->{height} - $s->{bottom_margin},
        $s->{axis_color}
    );

    $s->{image} = $image;
}

#
# Put hash marks and labels on the graph
#
sub label_x_axis($) {
    my $s = shift;

    my (@start_date, $tic_time, $tic_offset, $tic_val, $time,);

    if ($s->{type} eq 'today') {

        # Possibly put a label every hour (3600 seconds)
        for ($tic_time = $s->{x_start_time} ;
            $tic_time <= $s->{x_end_time} ; $tic_time += 3600)
        {
            $tic_val    = "";
            $tic_offset = ($tic_time - $s->{x_start_time});

            # Midnight?
            if ($tic_offset == 0 || $tic_offset == 86400) {
                $tic_val = "M";
            }
            elsif ($tic_offset % ($s->{x_hour_marks} * 3600) == 0) {
                if ($tic_offset == 43200) {
                    $tic_val = "Noon";
                }
                elsif ($tic_offset < 43200) {
                    $tic_val = $tic_offset / 3600 . "am";
                }
                else {
                    $tic_val = ($tic_offset - 43200) / 3600 . "pm";
                }
            }
            if (length($tic_val)) {
                $s->label_x_axis_point($tic_time, $tic_val, 1);
            }
        }
    }
    elsif ($s->{type} eq 'week') {

        # Put tics every 12 hours; label them at noon
        for (
            $tic_time = $s->{x_start_time} ;
            $tic_time <= $s->{x_end_time} ;
            $tic_time += 43200
          )
        {
            $tic_offset = ($tic_time - $s->{x_start_time});
            if ($tic_offset % 86400 == 0) {
                $s->label_x_axis_point($tic_time, "", 1);
            }
            elsif ($tic_offset % 43200 == 0) {
                @start_date = split /\s+/, localtime($tic_time);
                $tic_val = $start_date[1] . " " . $start_date[2];
                $s->label_x_axis_point($tic_time, $tic_val, 0);
            }
        }
    }
    elsif ($s->{type} eq 'month') {

        # Put tics every 12 hours; label them at noon
        for (
            $tic_time = $s->{x_start_time} ;
            $tic_time <= $s->{x_end_time} ;
            $tic_time += 43200
          )
        {
            $tic_offset = ($tic_time - $s->{x_start_time});
            if ($tic_offset % 86400 == 0) {
                $s->label_x_axis_point($tic_time, "", 1);
            }
            elsif ($tic_offset % 43200 == 0) {
                @start_date = split /\s+/, localtime($tic_time);
                $tic_val = $start_date[2];
                $s->label_x_axis_point($tic_time, $tic_val, 0);
            }
        }
    }
    elsif ($s->{type} eq 'year') {

        # Put tics at the first of each month; label them on the 15th
        # This is really inefficient, but we have to go 1 hour at a time
        # because daylight savings time can throw us off.
        for ($tic_time = $s->{x_start_time} ;
            $tic_time <= $s->{x_end_time} ; $tic_time += 3600)
        {
            $tic_offset = ($tic_time - $s->{x_start_time});
            if ($tic_offset == 0) {
                $s->label_x_axis_point($tic_time, "", 1);
            }
            else {
                $time = localtime($tic_time);
                $time =~ tr/:/ /;
                @start_date = split /\s+/, $time;
                if ($start_date[2] == 1 && $start_date[3] == 12) {
                    $s->label_x_axis_point($tic_time, "", 1);
                }
                elsif ($start_date[2] == 15 && $start_date[3] == 12) {
                    $tic_val = substr($start_date[1], 0, 3);
                    $s->label_x_axis_point($tic_time, $tic_val, 0);
                }
            }
        }

        # Put a tic at the end of the x-axis
        $s->label_x_axis_point($tic_time, "", 1);
    }
}

#
# Label a point on the x-axis
#
sub label_x_axis_point($$$$) {
    my $s = shift;

    my ($tic_time, $tic_val, $put_hash_mark) = @_;

    my $x_val = $s->{x_origin} + (($tic_time - $s->{x_start_time}) / $s->{x_scale});
    my $y_val = ($s->{height} - $s->{bottom_margin}) + $s->{hash_length};

    if ($put_hash_mark) {
        $s->{image}->line($x_val, $s->{y_origin}, $x_val, $s->{y_origin} + $s->{hash_length},
            $s->{axis_color});
    }

    if (length($tic_val)) {

        # Put the label on the graph. Center the text around the point
        $x_val =
          $s->{x_origin} + (($tic_time - $s->{x_start_time}) / $s->{x_scale}) -
          ((length($tic_val) * $s->{label_font}->width) / 2);
        $s->{image}->string($s->{label_font}, $x_val, $y_val, $tic_val, $s->{axis_color});
    }
}

#
# Label the y-axis
#
sub label_y_axis($) {
    my $s = shift;

    my ($delta, $val);

    # There are 5 hash marks on the y-axis; calculate the difference
    # between them.
    $delta = ($s->{max_y_val} / 5);

    for ($val = 0 ; $val <= $s->{max_y_val} ; $val += $delta) {
        my $y_val = ($s->{height} - $s->{bottom_margin}) - ($val / $s->{y_scale});

        # Put the label on the graph
        my $text = sprintf("%3d ", $val);
        my $title_x =
          $s->{x_origin} - $s->{hash_length} - (length($text) * $s->{legend_font}->width);
        $s->{image}
          ->string($s->{legend_font}, $title_x, $y_val - ($s->{legend_font}->height / 2),
            $text, $s->{axis_color});

        # Draw the hash mark
        $s->{image}->line($s->{x_origin} - $s->{hash_length},
            $y_val, $s->{x_origin}, $y_val, $s->{axis_color});
    }
}

#
# Process the license samples
#
sub load_license_samples($$) {
    my ($s, $infile) = @_;
    my $sample = {
        min_sample_time      => 0,
        max_sample_time      => 0,
        max_used_this_period => 0,
        max_used_ever        => 0,
        max_licenses_avail   => 0
    };
    my (%avail, %used);

    # Open the file
    local *SAMPLE;
    open SAMPLE, $infile or die "Can't open $infile: $!\n";

    # process the samples
    while (<SAMPLE>) {

        # Skip comments and whitespace
        next if (m/^#/);
        next if (m/^\s*$/);

        my ($sample_time, $lic_avail, $lic_used) = split ' ';

        # Is this sample within the graph's time period?
        if ($sample_time >= $s->{x_start_time} && $sample_time <= $s->{x_end_time}) {

            # Save the minimum and maximum sample times
            $sample->{min_sample_time} = $sample_time
              if ( $sample->{min_sample_time} == 0
                or $sample_time < $sample->{min_sample_time});

            $sample->{max_sample_time} = $sample_time
              if $sample_time > $sample->{max_sample_time};

            # Is there a new maximum for this graph?
            $sample->{max_used_this_period} = $lic_used
              if $lic_used > $sample->{max_used_this_period};

            $avail{$sample_time} = $lic_avail;
            $used{$sample_time}  = $lic_used;

        }

        # Is there new overall maximums?
        $sample->{max_used_ever} = $lic_used
          if $lic_used > $sample->{max_used_ever};
        $sample->{max_licenses_avail} = $lic_avail
          if $lic_avail > $sample->{max_licenses_avail};
    }
    close SAMPLE;

    die "No valid samples were found for TestTrack\n"
      if not $sample->{max_licenses_avail};

    $sample->{avail} = \%avail;
    $sample->{used}  = \%used;

    $s->{sample} = $sample;
}

1;

