# This module is part of the license usage graph package. It reads
# the number of liceses in use from a TestTrack Pro SOAP service.

package TTProLicenses;

use Tie::File;
use SOAP::Lite;

# Create a new object with required values for a user, password,
# and logfile.
sub new {
    my ($class, %values) = @_;

    die "TTProLicenses: must define user, password and log\n"
        if (
            not exists $values{user} or
            not exists $values{password} or
            not exists $values{logfile}
        );
    my $retval = {
        user     => $values{user},
        password => $values{password},
        logfile  => $values{logfile},
        wsdl     => $values{wsdl}
    };
    $retval->{host} = $values{host} || "http://localhost/cgi-bin/ttsoapcgi.exe";

    # The default wsdl file is located in the root of the host url
    if (not $retval->{wsdl}) {
        ($retval->{wsdl}) = $values{host} =~ m{^(http://[^/]+)};
        $retval->{wsdl} .= '/ttsoapcgi.wsdl';
    }
    
    bless $retval, $class;
    return $retval;
}

# Subroutine using TestTrack SOAP to count the number of available and
# used licenses
sub collect {
    my $o = shift;
   
    # Variables
    my $ns = 'urn:testtrack-interface';
    my $schema = 'http://www.w3.org/2001/XMLSchema';
    
    # The $soap object is our connection to the web services server
    my $soap = SOAP::Lite->service($o->{wsdl})->xmlschema($schema);
    
    # Get the list of databases, and the first database
    my $dbList = $soap->getDatabaseList();
    my $db = shift @{$dbList};
    
    # Login to the first database
    my $cookie = $soap->DatabaseLogon($db->{name}, $o->{user}, $o->{password});
    if ($soap->call->fault) {
        my $faultstring = $soap->call->faultstring;
        die
            "$0: login error: $faultstring\n" .
            "Cannot access " . $db->{name} . " at $dbHost";
    }
    
    # Retrieve the license list
    my $list = $soap->getUserLicenseList($cookie);
    if ($soap->call->fault) {
        my $faultstring = $soap->call->faultstring;
        $soap->DatabaseLogoff($cookie);
        die "$0: get license error: $faultstring\n";
    }
    
    # Count the used and total licenses
    my ($used, $total);
    for my $license (@{$list}) {
       $used  += $license->{numberUsed};
       $total += $license->{numberTotal};
    }
    if (not $total) {
        $soap->DatabaseLogoff($cookie);
        die "$0: no licenses found in $db->{name}\n";
    }
 
    # Write to the log file
    my $now = time;
    tie my @log, 'Tie::File', $o->{logfile};
    push @log, "$now $total $used " . localtime($now);
    # cut the log file down to size
    # splice @log, 0, @log - $max_log_size if @log > $max_log_size;
    untie @log;
    
    # Have to make sure to logoff
    $soap->DatabaseLogoff($cookie);
}

1;
